// This file is part of the FlexVer™ reference implementation
//
// © 2017 - 2018 Raptor Engineering, LLC
// All Rights Reserved
//
// WARNING
// This material may be covered by patent(s) and/or patent(s) pending
//
// Any use of this file without an active FlexVer™ license from
// Raptor Engineering, LLC is strictly prohibited.  A single
// exception to this is made for the sole purpose of third-party
// security auditing of FlexVer™ and associated technologies.
//
// For the security audit exception to apply, the primary purpose
// of the use must be to audit the security of the FlexVer™ technology,
// and Raptor Engineering, LLC must be notified in writing of
// any security flaw(s) or potential security flaw(s) found during
// the audit no less than four (4) weeks in advance of public
// disclosure of said flaw(s).  Furthermore, the number of compiled
// FlexVer™ binaries must be kept to the minimum necessary to conduct
// the security audit, and all compiled FlexVer™ binaries must be
// destroyed once the audit has been completed.  These restrictions
// do not apply to organizations or individuals possessing an active
// FlexVer™ license from Raptor Engineering, LLC.
//
// Additional license(s) beyond an active FlexVer™ license may be required
// for specific use cases.  Contact Raptor Engineeing, LLC for details.
//
// For sales and licensing information, please visit
// https://www.raptorengineering.com
//
// PUBLIC VERSION FOR GENERAL REDISTRIBUTION
// Licensed for use with Raptor Data Solutions, LLC's IntegriCloud™ service

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>
#include <sys/io.h>
#include <sys/ioctl.h>
#include <ifaddrs.h>
#include <netdb.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <curl/curl.h>

#include <b64.h>
#include <sha2.h>
#include <jsmn.h>
#include <uECC.h>
#include <libflexver.h>

#include "mticlient.h"

#define VERSION "0.9.1"

// Uncomment this to enable extra FlexVer™ debugging
// #define DEBUG_LIBFLEXVER 1

#define CERTIFICATION_TEXT	"I, %s, do hereby certify that I have personally inspected the FlexVer™ system with public key %s on %s, and that to the best of my knowledge this FlexVer™ system has not been not tampered with and furthermore that this FlexVer™ system is programmed with a firmware image whose SHA256 hash is %s.  %s%s%sI further declare, under penalty of perjury, that I have not been coerced into falsely making or signing this statement.  I hereby certify these statements to be true and correct by signing this message with my digital signature.\n"
#define CERT_NOTES_TEXT_HEADER	"As a condition of making this statement, I note for the record that this FlexVer™ system has the following properties: "
#define CERT_NOTES_TEXT_FOOTER	".  "
#define CERT_NO_NOTES		"As a condition of making this statement, I note for the record that I am not aware of any abnormalities or unusual properties associated with this FlexVer™ system.  "

#define MTISHELL_USAGE_STRING	"Usage: ./mticlient [-t] -u <username> <VM name> | ./mticlient [-t] -u <username> -i <VM ID> | ./mticlient [-t] -n <machine ID>"

#define PGP_VERIFY_COMMAND	"gpg --verify %s %s 2>&1"
#define SSH_KNOWN_HOSTS_FORMAT	"%s%s ssh-rsa %s\n"
#define SSH_ENVIRONMENT_FORMAT	"FV_AUTH_PK=\"%s\" FV_PROD_KEYLIST=\"%s\" FV_AUTH_FW_HASH=\"%s\" FV_PROD_FW_HASHLIST=\"%s\" FV_AUTH_MTI_HASH=\"%s\" FV_PROD_MTI_HASHLIST=\"%s\" FV_AUTH_AUTOSET_SESSION_KEY_AND_HASH_LISTS=\"%s\" "
#define SSH_COMMAND_FORMAT	"%ssshpass -f %s ssh -q -o \"SendEnv=FV_AUTH_PK\" -o \"SendEnv=FV_PROD_KEYLIST\" -o \"SendEnv=FV_AUTH_FW_HASH\" -o \"SendEnv=FV_PROD_FW_HASHLIST\" -o \"SendEnv=FV_AUTH_MTI_HASH\" -o \"SendEnv=FV_AUTH_AUTOSET_SESSION_KEY_AND_HASH_LISTS\" -o \"SendEnv=FV_PROD_MTI_HASHLIST\" -o \"SendEnv=MACHINE_SHELL\" -o \"StrictHostKeyChecking=yes\" -o \"UserKnownHostsFile=%s\" -p %s %s@%s"

#define POST_REQUEST_BUFFER_SIZE	4096
#define FLEXVER_PUBLIC_KEY_LENGTH	64
#define SYSTEM_FIRMWARE_HASH_LENGTH	32
#define MTI_FIRMWARE_HASH_LENGTH	32

#define PRODUCTION_FV_PK_LIST_FILE_1	"/.integricloud/production_flexver_hashes.txt"
#define PRODUCTION_FV_PK_LIST_FILE_2	"/etc/integricloud/production_flexver_hashes.txt"
#define PRODUCTION_SYS_FW_LIST_FILE_1	"/.integricloud/production_system_firmware_hashes.txt"
#define PRODUCTION_SYS_FW_LIST_FILE_2	"/etc/integricloud/production_system_firmware_hashes.txt"
#define PRODUCTION_MTI_FW_LIST_FILE_1	"/.integricloud/production_mti_firmware_hashes.txt"
#define PRODUCTION_MTI_FW_LIST_FILE_2	"/etc/integricloud/production_mti_firmware_hashes.txt"

#define CONSOLE_NORMAL		"\x1B[0m"

#define CONSOLE_FONT_BLACK	"\x1B[30m"
#define CONSOLE_FONT_RED	"\x1B[31m"
#define CONSOLE_FONT_GREEN	"\x1B[32m"
#define CONSOLE_FONT_YELLOW	"\x1B[33m"
#define CONSOLE_FONT_BLUE	"\x1B[34m"
#define CONSOLE_FONT_MAGENTA	"\x1B[35m"
#define CONSOLE_FONT_CYAN	"\x1B[36m"
#define CONSOLE_FONT_WHITE	"\x1B[37m"
#define CONSOLE_FONT_NONE	""

#define CONSOLE_BG_BLACK	"\x1B[40m"
#define CONSOLE_BG_RED		"\x1B[41m"
#define CONSOLE_BG_GREEN	"\x1B[42m"
#define CONSOLE_BG_YELLOW	"\x1B[43m"
#define CONSOLE_BG_BLUE		"\x1B[44m"
#define CONSOLE_BG_MAGENTA	"\x1B[45m"
#define CONSOLE_BG_CYAN		"\x1B[46m"
#define CONSOLE_BG_WHITE	"\x1B[47m"
#define CONSOLE_BG_NONE		""

#define CONSOLE_STEADY		""
#define CONSOLE_BLINK		"\x1B[5m"

const char* control_server_api_url = "http://api.integricloud.com/v1/api";
const char* audit_log_url = "http://api.integricloud.com/v1/audit-log";

const char ssh_rsa_pubkey_preamble[11] = {0x00, 0x00, 0x00, 0x07, 0x73, 0x73, 0x68, 0x2d, 0x72, 0x73, 0x61};

char ssh_known_hosts_filename[] = "/tmp/mticlient_known_hosts_XXXXXX";
int ssh_known_hosts_fd = -1;
char ssh_password_filename[] = "/tmp/mticlient_password_XXXXXX";
int ssh_password_fd = -1;

char* production_fv_pk_list_file_1_expanded = NULL;
char* production_fv_pk_list_file_2_expanded = NULL;
char* production_sys_fw_list_file_1_expanded = NULL;
char* production_sys_fw_list_file_2_expanded = NULL;
char* production_mti_fw_list_file_1_expanded = NULL;
char* production_mti_fw_list_file_2_expanded = NULL;

uint8_t remote_hint_trust_client_keys = 0;

CURL *curl;
CURLcode curl_ret;

jsmn_parser json_parser;
jsmntok_t json_tokens[128];	// No more than 128 tokens should be returned at any given time
char* json_string = 0;
int json_parser_token_count = 0;
int json_parser_start_token = 1;
int json_valid_api_response = 0;

void mticlient_cleanup() {
	if (ssh_password_fd != -1) {
		unlink(ssh_password_filename);
		close(ssh_password_fd);
		ssh_password_fd = -1;
	}

	if (ssh_known_hosts_fd != -1) {
		unlink(ssh_known_hosts_filename);
		close(ssh_known_hosts_fd);
		ssh_known_hosts_fd = -1;
	}
}

mticlient_log_level mticlient_debug_level = MTICLIENT_LOGLEVEL_NONE;

void mticlient_log(mticlient_log_level log_level, const char* format, ...) {
	va_list args;
	va_start(args, format);
	if (mticlient_debug_level >= log_level) {
		vprintf(format, args);
	}
	va_end(args);
}

void set_mticlient_loglevel(mticlient_log_level log_level) {
	mticlient_debug_level = log_level;
}

// NOTE
// The returned string must be freed with free() after use!
char* get_host_ip_addresses(const char* hostname, uint16_t port) {
	int ret;
	char* realloc_ptr = NULL;
	char addr_buffer[1024+1];
	const char* addr_ptr = NULL;
	struct addrinfo hints;
	struct addrinfo* rhost_info;
	struct addrinfo* result;
	struct sockaddr_in* addr_info;
	char* result_buffer = NULL;
	size_t result_buffer_pos = 0;
	size_t result_buffer_length = 0;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_UNSPEC;		// Look up both IPv4 and IPv6 addresses
	hints.ai_socktype = SOCK_STREAM;	// TCP connection desired

	ret = getaddrinfo(hostname, "ssh", &hints, &rhost_info);
	if (ret) {
		mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Unable to resolve remote node address!  Details: '%s'\n", gai_strerror(ret));
		return NULL;
	}

	for (result = rhost_info; result; result = result->ai_next) {
		if ((result->ai_family == AF_INET) || (result->ai_family == AF_INET6)) {
			addr_info = (struct sockaddr_in*)result->ai_addr;
			if (addr_info) {
				memset(addr_buffer, 0, sizeof(addr_buffer));
				addr_ptr = inet_ntop(result->ai_family, &addr_info->sin_addr, addr_buffer, sizeof(addr_buffer));
				if (addr_ptr) {
					addr_buffer[1024] = 0;
					if (port == 22) {
						result_buffer_length += snprintf(NULL, 0, ",%s", addr_buffer);
					}
					else {
						result_buffer_length += snprintf(NULL, 0, ",[%s]:%d", addr_buffer, port);
					}
					result_buffer_length++;
					realloc_ptr = realloc(result_buffer, result_buffer_length);
					if (realloc_ptr) {
						result_buffer = realloc_ptr;
						if (port == 22) {
							snprintf(result_buffer + result_buffer_pos, result_buffer_length - result_buffer_pos, ",%s", addr_buffer);
						}
						else {
							snprintf(result_buffer + result_buffer_pos, result_buffer_length - result_buffer_pos, ",[%s]:%d", addr_buffer, port);
						}
						result_buffer_pos = result_buffer_length - 1;
					}
					else {
						mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Unable to allocate memory for remote node address storage!\n");
						break;
					}
				}
			}
		}
	}
	freeaddrinfo(rhost_info);

	return result_buffer;
}

int json_skip_next_token(int start_token) {
	int i;
	int j = start_token;
	for (i = 0; i < json_tokens[start_token].size; i++) {
		j += json_skip_next_token(j+1);
	}
	return (j - start_token) + 1;
}

// NOTE
// The returned string must be freed with free() after use!
uint8_t contains_key_value(int start_token, const char * search_key) {
	int i;
	size_t string_length = 0;

	for (i = start_token; i < json_parser_token_count; i++) {
		if (jsoneq(json_string, &json_tokens[i], search_key) == 0) {
			return 1;
		}
	}

	return 0;
}

// NOTE
// The returned string must be freed with free() after use!
char* find_key_value(int start_token, const char * search_key) {
	int i;
	size_t string_length = 0;
	char* return_value = NULL;

	for (i = start_token; i < json_parser_token_count; i++) {
		if (jsoneq(json_string, &json_tokens[i], search_key) == 0) {
			string_length = json_tokens[i+1].end - json_tokens[i+1].start;
			return_value = malloc(string_length + 1);
			if (!return_value) {
				return NULL;
			}
			memset(return_value, 0, string_length + 1);
			strncpy(return_value, json_string + json_tokens[i+1].start, string_length);
			return_value[string_length] = 0;
			return return_value;
		}
	}

	return NULL;
}

// NOTE
// The returned string must be freed with free() after use!
char* extract_key_value(int start_token, const char* buffer) {
	size_t string_length = 0;
	char* return_value = NULL;

	string_length = json_tokens[start_token+1].end - json_tokens[start_token+1].start;
	return_value = malloc(string_length + 1);
	if (!return_value) {
		return NULL;
	}
	memset(return_value, 0, string_length + 1);
	strncpy(return_value, buffer + json_tokens[start_token+1].start, string_length);
	return_value[string_length] = 0;
	return return_value;
}

int detect_and_process_json_error_response(int start_token, const char* return_type_value) {
	char* json_value = 0;

	if (strcmp(return_type_value, "ERROR") == 0) {
		json_value = find_key_value(json_parser_start_token, "returnData");
		if (json_value) {
			if (strcmp(json_value, "SYSTEM_BUSY") == 0) {
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Remote system busy, retrying...\n");
				json_valid_api_response = 2;
			}
			else {
				mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Remote server returned ERROR response!  Details: '%s'\n", json_value);
				json_valid_api_response = 0;
			}
			free(json_value);
			json_value = NULL;
		}
		else {
			json_valid_api_response = 0;
		}
		return 1;
	}

	json_valid_api_response = 1;
	return 0;
}

uint8_t* curl_buffer = NULL;
size_t curl_buffer_size = 0;

static size_t mticlient_curl_receive_data(void* buffer, size_t size, size_t nmemb, void* arg) {
	uint32_t curl_buffer_orig_size;
	void* realloc_ptr = NULL;

	if (!curl_buffer) {
		curl_buffer_orig_size = 0;
		curl_buffer_size = nmemb * size;
		curl_buffer = malloc(curl_buffer_size);
	}
	else {
		curl_buffer_orig_size = curl_buffer_size;
		curl_buffer_size += nmemb * size;
		realloc_ptr = realloc(curl_buffer, curl_buffer_size);
		if (realloc_ptr) {
			curl_buffer = realloc_ptr;
			realloc_ptr = NULL;
		}
		else {
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
			return 0;
		}
	}

	memcpy(curl_buffer + curl_buffer_orig_size, buffer, nmemb * size);

	return nmemb * size;
}

int initialize_curl() {
	if (curl) {
		return 1;
	}

	// Initialise cURL and get handle
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();
	if (!curl) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to initialize cURL!\n");
		return 1;
	}
	curl_easy_setopt(curl, CURLOPT_NOPROGRESS, 1L);
	curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);
	curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, mticlient_curl_receive_data);

	// Prevent continually increasing FD use at the possible expense of some performance
	curl_easy_setopt(curl, CURLOPT_FORBID_REUSE, 1L);

	return 0;
}

void tear_down_curl() {
	curl_global_cleanup();
	curl = NULL;
}

void mticlient_reset_curl_data_buffer() {
	free(curl_buffer);
	curl_buffer = NULL;
	curl_buffer_size = 0;
}

void mticlient_process_received_data() {
	int i;
	char* json_value = 0;

	curl_buffer_size++;
	curl_buffer = realloc(curl_buffer, curl_buffer_size);
	curl_buffer[curl_buffer_size-1] = 0;

	mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Received response '%s' from remote server\n", curl_buffer);

	jsmn_init(&json_parser);
	json_parser_token_count = jsmn_parse(&json_parser, curl_buffer, strlen(curl_buffer), json_tokens, sizeof(json_tokens) / sizeof(json_tokens[0]));
	if (json_parser_token_count < 0) {
		mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Unable to parse remote server response!  Return code: %d\n", json_parser_token_count);
		json_valid_api_response = 0;
	}
	else {
		if ((json_parser_token_count < 1) || (json_tokens[0].type != JSMN_OBJECT)) {
			mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Received invalid JSON from remote server!  Details: 'Top level object absent'\n");
		}
		else {
			free(json_string);
			json_string = malloc(curl_buffer_size);
			if (json_string) {
				memcpy(json_string, curl_buffer, curl_buffer_size);
				json_value = find_key_value(json_parser_start_token, "returnType");
				if (json_value) {
					detect_and_process_json_error_response(json_parser_start_token, json_value);
					free(json_value);
					json_value = NULL;
				}
			}
			else {
				mticlient_reset_curl_data_buffer();
				return;
			}
		}
	}

	mticlient_reset_curl_data_buffer();
}

void sanitize_command_buffer(char* str) {
	size_t i = 0;
	size_t string_length = 0;

	string_length = strlen(str);
	for (i = 0; i < string_length; i++) {
		if (((str[i] >= 0x30) && (str[i] <= 0x39))
			|| ((str[i] >= 0x41) && (str[i] <= 0x5a))
			|| ((str[i] >= 0x61) && (str[i] <= 0x7a))
			|| (str[i] == 0x2d)
			|| (str[i] == 0x2e)
			|| (str[i] == 0x5f)) {
			continue;
		}
		str[i] = '_';
	}
	str[string_length] = 0;
}

void free_datalist(char** data_list, size_t data_count) {
	size_t data_index;

	for (data_index = 0; data_index < data_count; data_index++) {
		free(data_list[data_index]);
		data_list[data_index] = 0;
	}
	free(data_list);
}

void free_hex_datalist(uint8_t** data_list, size_t data_count) {
	size_t data_index;

	for (data_index = 0; data_index < data_count; data_index++) {
		free(data_list[data_index]);
		data_list[data_index] = 0;
	}
	free(data_list);
}

// NOTE
// The returned string must be freed with free_hex_datalist() after use!
int read_hex_datalist_from_file(const char* filename, uint8_t*** data_list, size_t* data_count, size_t element_length) {
	FILE* vm_data_fp = NULL;
	size_t pos;
	size_t len = 0;
	ssize_t read;
	ssize_t data_pos = 0;
	ssize_t data_length = 0;
	uint8_t data_value = 0;
	char* line = NULL;
	char* endptr = NULL;
	char buffer[3] = { 0 };
	void* realloc_ptr = NULL;

	*data_count = 0;
	vm_data_fp = fopen(filename, "r");
	if (!vm_data_fp) {
		mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] File '%s' could not be opened for read!\n", filename);
		return 1;
	}

	while (read = getline(&line, &len, vm_data_fp) != -1) {
		(*data_count)++;
		realloc_ptr = realloc(*data_list, sizeof(uint8_t*) * (*data_count));
		if (!realloc_ptr) {
			free_hex_datalist(*data_list, (*data_count) - 1);
			*data_list = NULL;
			*data_count = 0;
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
			return 1;
		}
		*data_list = realloc_ptr;
		realloc_ptr = NULL;
		pos = strlen(line) - 1;
		while (line[pos] == '\n') {
			pos--;
		}
		line[pos+1] = 0;

		data_length = strlen(line);
		(*data_list)[(*data_count)-1] = malloc(element_length + 1);
		if (!((*data_list)[(*data_count)-1])) {
			free_hex_datalist(*data_list, (*data_count) - 1);
			*data_list = NULL;
			*data_count = 0;
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
			return 1;
		}

		memset((*data_list)[(*data_count)-1], 0, element_length);
		for (data_pos = 0; data_pos < element_length; data_pos++) {
			if (((data_pos * 2) + 1) >= data_length) {
				break;
			}
			buffer[0] = line[data_pos * 2];
			buffer[1] = line[(data_pos * 2) + 1];
			buffer[2] = 0;
			data_value = strtol(buffer, &endptr, 16);
			if ((*buffer == 0) || (endptr && (*endptr != 0))) {
				mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to parse data value '%s'!\n", buffer);
				(*data_list)[(*data_count)-1][data_pos] = 0;
			}
			else {
				(*data_list)[(*data_count)-1][data_pos] = data_value;
			}
		}
	}
	free(line);
	fclose(vm_data_fp);

	return 0;
}

// NOTE
// The returned string must be freed with free() after use!
char* write_raw_datalist_to_b64_string(uint8_t** data_list, const size_t data_count, size_t element_length) {
	char* ret_list = NULL;
	char* data_encoded = NULL;
	size_t ret_list_len = 0;
	size_t ret_list_pos = 0;
	size_t data_index = 0;
	size_t data_length = 0;

	// Assemble data list
	ret_list_len = 1;
	for (data_index = 0; data_index < data_count; data_index++) {
		data_encoded = b64_encode(data_list[data_index], element_length);
		if (!data_encoded) {
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
			return NULL;
		}
		ret_list_len += strlen(data_encoded) + 1;
		free(data_encoded);
	}
	if (ret_list_len < 2) {
		return strdup("");
	}
	ret_list = malloc(ret_list_len);
	if (!ret_list) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return NULL;
	}
	memset(ret_list, 0, ret_list_len);
	ret_list_pos = 0;
	if (data_count > 0) {
		for (data_index = 0; data_index < data_count; data_index++) {
			data_encoded = b64_encode(data_list[data_index], element_length);
			if (data_encoded) {
				data_length = strlen(data_encoded);
				memcpy(ret_list + ret_list_pos, data_encoded, sizeof(char) * data_length);
				free(data_encoded);
				ret_list_pos += data_length;
				ret_list[ret_list_pos] = ' ';
				ret_list_pos++;
			}
		}
		ret_list[ret_list_pos-1] = 0;
		ret_list[ret_list_len-1] = 0;
	}
	else {
		ret_list[0] = 0;
	}

	return ret_list;
}

// NOTE
// The returned string must be freed with free() after use!
char* raw_data_to_hex_string(uint8_t* data, size_t data_length) {
	char* ret_string = NULL;
	size_t data_pos = 0;

	ret_string = malloc((data_length * 2) + 1);
	if (!ret_string) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return NULL;
	}

	for (data_pos = 0; data_pos < data_length; data_pos++) {
		snprintf(ret_string + (data_pos * 2), 3, "%02x", data[data_pos]);
	}
	ret_string[(data_length * 2)] = 0;

	return ret_string;
}

// NOTE
// The returned string must be freed with free() after use!
char* write_datalist_to_string(char** data_list, const size_t data_count) {
	char* ret_list = NULL;
	size_t ret_list_len = 0;
	size_t ret_list_pos = 0;
	size_t data_index = 0;
	size_t data_length = 0;

	// Assemble data list
	ret_list_len = 1;
	for (data_index = 0; data_index < data_count; data_index++) {
		ret_list_len += strlen(data_list[data_index]) + 1;
	}
	if (ret_list_len < 2) {
		return strdup("");
	}
	ret_list = malloc(ret_list_len);
	if (!ret_list) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return NULL;
	}
	memset(ret_list, 0, ret_list_len);
	ret_list_pos = 0;
	if (data_count > 0) {
		for (data_index = 0; data_index < data_count; data_index++) {
			data_length = strlen(data_list[data_index]);
			memcpy(ret_list + ret_list_pos, data_list[data_index], sizeof(char) * data_length);
			ret_list_pos += data_length;
			ret_list[ret_list_pos] = ' ';
			ret_list_pos++;
		}
		ret_list[ret_list_pos-1] = 0;
		ret_list[ret_list_len-1] = 0;
	}
	else {
		ret_list[0] = 0;
	}

	return ret_list;
}

// NOTE
// The returned strings must be freed with free() after use!
int get_node_id_from_vm_info(const char* vm_id, const char* username, char** node_id, uint8_t id_lookup) {
	int i;
	int j;
	int line;
	int cur_pos;
	int cur_obj_end_token;
	int vm_list_start_token;
	int vm_list_end_token;
	char post_request_data[POST_REQUEST_BUFFER_SIZE+1];
	char* json_value = 0;
	char** realloc_ptr = NULL;
	char* json_key_value = NULL;
	uint8_t node_id_received = 0;

	// Initialize variables
	if (node_id) {
		*node_id = NULL;
	}

	// Send disk status request to control server
	if (id_lookup) {
		snprintf(post_request_data, POST_REQUEST_BUFFER_SIZE+1, "{\"command\":\"REQUEST_PARAMETER\",\"parameter\":\"EXTERNAL_NODE_ID_FOR_VM_ID\",\"vmID\":\"%s\",\"username\":\"%s\"}", vm_id, username);
	}
	else {
		snprintf(post_request_data, POST_REQUEST_BUFFER_SIZE+1, "{\"command\":\"REQUEST_PARAMETER\",\"parameter\":\"EXTERNAL_NODE_ID_FOR_VM_NAME\",\"vmName\":\"%s\",\"username\":\"%s\"}", vm_id, username);
	}
	post_request_data[POST_REQUEST_BUFFER_SIZE] = 0;
	curl_easy_setopt(curl, CURLOPT_URL, control_server_api_url);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_request_data);
	mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending request '%s' to control server...\n", post_request_data);
	mticlient_reset_curl_data_buffer();
	curl_ret = curl_easy_perform(curl);
	if (curl_ret != CURLE_OK) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to retrieve node ID from control server!  Detail: %s\n", curl_easy_strerror(curl_ret));
		return 1;
	}
	else {
		mticlient_process_received_data();
		if (json_valid_api_response) {
			// Iterate over top level array
			vm_list_start_token = -1;
			vm_list_end_token = -1;
			for (i = 1; i < json_parser_token_count; i++) {
				// Skip all child objects
				if (json_tokens[i].type == JSMN_OBJECT) {
					i += json_skip_next_token(i) - 1;
					continue;
				}

				if (node_id) {
					if ((jsoneq(json_string, &json_tokens[i], "nodeID") == 0) && (json_tokens[i+1].type == JSMN_STRING)) {
						json_key_value = extract_key_value(i, json_string);
						if (json_key_value) {
							*node_id = strdup(json_key_value);
							free(json_key_value);
							if (!(*node_id)) {
								mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
								return 1;
							}
							node_id_received = 1;
						}
						else {
							mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to retrieve node ID from control server!  Detail: 'nodeID' key could not be parsed\n");
							return 1;
						}
					}
				}
				i++;
			}

			if (node_id) {
				if (!node_id_received) {
					mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to retrieve node ID from control server!  Detail: 'nodeID' key missing\n");
					return 1;
				}
			}
		}
		else {
			return 1;
		}
	}

	if (node_id && (*node_id)) {
		sanitize_command_buffer(*node_id);
	}

	return 0;
}

// NOTE
// The returned string list must be freed with free_datalist() after use!
int mticlient_process_audit_log(const char* search_flexver_public_key, char*** key_list, size_t* key_count) {
	int i;
	int j;
	int line;
	int cur_pos;
	int cur_obj_end_token;
	char* json_value = 0;
	int audit_log_entries = 0;
	int audit_failed = 1;
	void* realloc_ptr = NULL;

	curl_buffer = realloc(curl_buffer, curl_buffer_size+1);
	curl_buffer[curl_buffer_size] = 0;

	mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Received response '%s' from remote server\n", curl_buffer);

	jsmn_init(&json_parser);
	json_parser_token_count = jsmn_parse(&json_parser, curl_buffer, strlen(curl_buffer), json_tokens, sizeof(json_tokens) / sizeof(json_tokens[0]));
	if (json_parser_token_count < 0) {
		mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Unable to parse audit log!  Return code: %d\n", json_parser_token_count);
		json_valid_api_response = 0;
	}
	else {
		if ((json_parser_token_count < 1) || (json_tokens[0].type != JSMN_OBJECT)) {
			mticlient_log(MTICLIENT_LOGLEVEL_WARN, "[WARN] Received invalid JSON from remote server!  Details: 'Top level object absent'\n");
		}
		else {
			// Iterate over top level array
			for (i = 1; i < json_parser_token_count; i++) {
				cur_pos = i + 1;
				cur_obj_end_token = cur_pos + json_skip_next_token(cur_pos);
				if (json_tokens[i].type == JSMN_OBJECT) {
					char* flexver_public_key = NULL;
					char* flexver_firmware_hash = NULL;
					char* certification_date = NULL;
					char* certification_notes = NULL;
					char* signer_name = NULL;
					char* gpg_signature = NULL;
					size_t gpg_signature_length = 0;

					// Iterate over second level array, ignoring all child objects
					for (j = 0; j < json_tokens[i].size; j++) {
						if ((json_tokens[cur_pos].type == JSMN_STRING) && (json_tokens[cur_pos+1].type == JSMN_STRING)) {
							if (!flexver_public_key && (jsoneq(curl_buffer, &json_tokens[cur_pos], "flexverPublicKey") == 0)) {
								flexver_public_key = extract_key_value(cur_pos, curl_buffer);
							}
							if (!flexver_firmware_hash && (jsoneq(curl_buffer, &json_tokens[cur_pos], "flexverFirmwareHash") == 0)) {
								flexver_firmware_hash = extract_key_value(cur_pos, curl_buffer);
							}
							if (!certification_date && (jsoneq(curl_buffer, &json_tokens[cur_pos], "certificationDate") == 0)) {
								certification_date = extract_key_value(cur_pos, curl_buffer);
							}
							if (!certification_notes && (jsoneq(curl_buffer, &json_tokens[cur_pos], "certificationNotes") == 0)) {
								certification_notes = extract_key_value(cur_pos, curl_buffer);
							}
							if (!signer_name && (jsoneq(curl_buffer, &json_tokens[cur_pos], "signerName") == 0)) {
								signer_name = extract_key_value(cur_pos, curl_buffer);
							}
							if (!gpg_signature && (jsoneq(curl_buffer, &json_tokens[cur_pos], "gpgSignature") == 0)) {
								gpg_signature = extract_key_value(cur_pos, curl_buffer);
							}
						}
						cur_pos += json_skip_next_token(cur_pos);
					}

					if (flexver_public_key && flexver_firmware_hash && certification_date && certification_notes && signer_name && gpg_signature) {
						if (strcmp(flexver_public_key, search_flexver_public_key) == 0) {
							char* buffer = NULL;
							size_t buffer_len;
							uint8_t have_notes = 0;
							char pgp_result[16][1024+1];
							int pgp_result_line = 0;
							int pgp_return_code = 1;
							if (strlen(certification_notes) > 0) {
								have_notes = 1;
							}
							if (have_notes) {
								buffer_len = snprintf(NULL, 0, CERTIFICATION_TEXT, signer_name, flexver_public_key, certification_date, flexver_firmware_hash, CERT_NOTES_TEXT_HEADER, certification_notes, CERT_NOTES_TEXT_FOOTER);
							}
							else {
								buffer_len = snprintf(NULL, 0, CERTIFICATION_TEXT, signer_name, flexver_public_key, certification_date, flexver_firmware_hash, CERT_NO_NOTES, "", "");
							}
							buffer_len++;
							buffer = malloc(buffer_len);
							if (buffer) {
								if (have_notes) {
									snprintf(buffer, buffer_len, CERTIFICATION_TEXT, signer_name, flexver_public_key, certification_date, flexver_firmware_hash, CERT_NOTES_TEXT_HEADER, certification_notes, CERT_NOTES_TEXT_FOOTER);
								}
								else {
									snprintf(buffer, buffer_len, CERTIFICATION_TEXT, signer_name, flexver_public_key, certification_date, flexver_firmware_hash, CERT_NO_NOTES, "", "");
								}
								buffer[buffer_len-1] = 0;

								// Decode signature
								uint8_t* gpg_signature_decoded = b64_decode_ex(gpg_signature, strlen(gpg_signature), &gpg_signature_length);

								// Create plaintext and signature files
								char plaintext_filename[] = "/tmp/mticlient_auditlog_XXXXXX";
								char signature_filename[] = "/tmp/mticlient_auditlog_XXXXXX";
								int plaintext_fd;
								int signature_fd;
								plaintext_fd = mkstemp(plaintext_filename);
								if (plaintext_fd >= 0) {
									// Write plaintext to file
									write(plaintext_fd, buffer, buffer_len-1);
									fsync(plaintext_fd);
									mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Wrote plaintext to file '%s'\n", plaintext_filename);

									signature_fd = mkstemp(signature_filename);
									if (signature_fd >= 0) {
										// Write signature to file
										write(signature_fd, gpg_signature_decoded, gpg_signature_length);
										fsync(signature_fd);
										mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Wrote signature to file '%s'\n", signature_filename);

										char* command_buffer = NULL;
										size_t command_buffer_len;

										// Verify signature with external PGP program
										memset(pgp_result, 0, sizeof(pgp_result));
										command_buffer_len = snprintf(NULL, 0, PGP_VERIFY_COMMAND, signature_filename, plaintext_filename);
										command_buffer_len++;
										command_buffer = malloc(command_buffer_len);
										if (command_buffer) {
											snprintf(command_buffer, command_buffer_len, PGP_VERIFY_COMMAND, signature_filename, plaintext_filename);
											command_buffer[command_buffer_len-1] = 0;
											mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Executing '%s' to verify signature...\n", command_buffer);
											FILE* pgp_verify_fp = popen(command_buffer, "r");
											if (pgp_verify_fp != NULL) {
												char gpg_output_unexpected = 0;
												while (fgets(pgp_result[pgp_result_line], sizeof(pgp_result[pgp_result_line])-1, pgp_verify_fp) != NULL) {
													pgp_result[pgp_result_line][strlen(pgp_result[pgp_result_line])-1] = 0;
													if ((pgp_result[pgp_result_line][0] == 'g')
														&& (pgp_result[pgp_result_line][1] == 'p')
														&& (pgp_result[pgp_result_line][2] == 'g')
														&& (pgp_result[pgp_result_line][3] == ':')
														&& (pgp_result[pgp_result_line][4] == ' ')) {
														pgp_result_line++;
													}
													if (pgp_result_line == (sizeof(pgp_result) / sizeof(pgp_result[0]))) {
														gpg_output_unexpected = 1;
														break;
													}
												}
												pgp_return_code = pclose(pgp_verify_fp);
												if (gpg_output_unexpected) {
													pgp_return_code = -1;
												}
											}
											free(command_buffer);
										}
										unlink(signature_filename);
										close(signature_fd);
									}
									unlink(plaintext_filename);
									close(plaintext_fd);
								}

								free(gpg_signature_decoded);
								free(buffer);
							}

							// Check FlexVer™ firmware hash against local list of production hashes
							uint8_t flexver_confirmed_production_hash = 0;
							char flexver_firmware_hash_comments[1024+1];
							FILE* flexver_hash_file_fp = fopen(production_fv_pk_list_file_1_expanded, "r");
							if (!flexver_hash_file_fp) {
								flexver_hash_file_fp = fopen(production_fv_pk_list_file_2_expanded, "r");
							}
							if (flexver_hash_file_fp) {
								size_t len = 0;
								size_t pos;
								ssize_t read;
								char* line = NULL;
								while (read = getline(&line, &len, flexver_hash_file_fp) != -1) {
									pos = strlen(line)-1;
									while (line[pos] == '\n') {
										pos--;
									}
									line[pos+1] = 0;
									if (strlen(line) > 0) {
										if (strcmp(line, flexver_firmware_hash) == 0) {
											flexver_confirmed_production_hash = 1;
										}
									}
								}
								free(line);
								fclose(flexver_hash_file_fp);
							}

							if (key_list && key_count) {
								if (flexver_confirmed_production_hash && (pgp_return_code == 0)) {
									realloc_ptr = realloc(*key_list, sizeof(char*) * ((*key_count) + 1));
									if (!realloc_ptr) {
										mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
									}
									else {
										*key_list = realloc_ptr;
										realloc_ptr = NULL;
										(*key_list)[*key_count] = strdup(flexver_public_key);
										if (!(*key_list)[*key_count]) {
											mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
										}
										else {
											(*key_count)++;
										}
									}
								}
							}

							memset(flexver_firmware_hash_comments, 0, sizeof(flexver_firmware_hash_comments));
							if (flexver_confirmed_production_hash) {
								if (pgp_return_code == 0) {
									snprintf(flexver_firmware_hash_comments, 1024, "%sVERIFIED%s", CONSOLE_FONT_GREEN, CONSOLE_NORMAL);
								}
								else {
									snprintf(flexver_firmware_hash_comments, 1024, "%sUNKNOWN%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
								}
							}
							else {
								snprintf(flexver_firmware_hash_comments, 1024, "%sUNVERIFIED%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
							}

							audit_log_entries++;
							if (pgp_return_code == 0) {
								mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] [ %sAUDIT%s ]  Auditor: %s\n[INFO]            Date: %s\n[INFO]            FlexVer™ firmware hash:   %s\n[INFO]            FlexVer™ firmware status: %s\n[INFO]            Notes: %s\n", CONSOLE_FONT_GREEN, CONSOLE_NORMAL, signer_name, certification_date, flexver_firmware_hash, flexver_firmware_hash_comments, certification_notes);
								audit_failed = 0;
							}
							else {
								mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] [ %sAUDIT%s ]  Auditor: %s\n[INFO]            Date: %s\n[INFO]            FlexVer™ firmware hash:   %s\n[INFO]            FlexVer™ firmware status: %s\n[INFO]            Notes: %s\n", CONSOLE_FONT_RED, CONSOLE_NORMAL, signer_name, certification_date, flexver_firmware_hash, flexver_firmware_hash_comments, certification_notes);
							}
							for (line = 0; line < pgp_result_line; line++) {
								mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]            %s\n", pgp_result[line]);
							}
						}
					}

					free(flexver_public_key);
					free(flexver_firmware_hash);
					free(certification_date);
					free(certification_notes);
					free(signer_name);
					free(gpg_signature);
				}
				i = cur_pos - 1;
			}
		}
	}

	if (audit_log_entries < 1) {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] [ %sAUDIT%s ]  No audit information available!\n", CONSOLE_FONT_RED, CONSOLE_NORMAL);
	}

	free(curl_buffer);
	curl_buffer = NULL;
	curl_buffer_size = 0;

	return audit_failed;
}

ssize_t command_argument_index(char** command_components, size_t command_component_count, const char* argument) {
	int i;

	for (i = 0; i < command_component_count; i++) {
		if (strcmp(command_components[i], argument) == 0) {
			return i;
		}
	}

	return -1;
}

int verify_remote_flexver_system(char* destination_machine_id, char* random_nonce_encoded, uint8_t flexver_verification_time_hint, uint8_t* verification_result, char** flexver_public_key, char** flexver_system_firmware_hash, char** flexver_mti_firmware_hash, char** flexver_signature, char** flexver_status, char** flexver_internal_boot_count, char** flexver_rsa_public_key, uint8_t** flexver_public_key_decoded, uint8_t** flexver_signature_decoded, uint32_t* flexver_status_raw, uint16_t* flexver_internal_boot_count_raw, uint8_t** flexver_system_firmware_hash_decoded, uint8_t**flexver_mti_firmware_hash_decoded, uint8_t** flexver_rsa_public_key_decoded, char** flexver_public_key_string_1, char** flexver_public_key_string_2, char** flexver_system_firmware_hash_string, char** flexver_mti_firmware_hash_string, uint8_t* random_nonce_raw, uint8_t* flexver_signature_valid, uint8_t pass) {
	char post_request_data[POST_REQUEST_BUFFER_SIZE+1];
	char* json_value = NULL;
	uint32_t elapsed_time_milliseconds = 0;
	char has_flexver = 1;

	// Request remote FlexVer™ authentication
	while (1) {
		snprintf(post_request_data, 512, "{\"command\":\"REQUEST_OPERATION\",\"operation\":\"FLEXVER_AUTHENTICATION\",\"machineID\":\"%s\",\"nonce\":\"%s\"}", destination_machine_id, random_nonce_encoded);
		post_request_data[512] = 0;
		curl_easy_setopt(curl, CURLOPT_URL, control_server_api_url);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_request_data);
		mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending request '%s' to remote server...\n", post_request_data);
		mticlient_reset_curl_data_buffer();
		curl_ret = curl_easy_perform(curl);
		if (curl_ret != CURLE_OK) {
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to request FlexVer™ verification on remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
			return 1;
		}
		else {
			mticlient_process_received_data();
		}
		if ((json_valid_api_response == 0) || (json_valid_api_response == 1)) {
			// Break out of loop on either hard error or success
			break;
		}
		sleep(1);
	}

	if (flexver_verification_time_hint > 0) {
		if (pass == 1) {
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Integrity verification may take up to %d seconds to complete, please wait...\n", flexver_verification_time_hint);
		}
		else {
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "\n");
		}
	}
	else {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] %sFlexVer™ verification not available for this node (FlexVer™ may not be installed) %s\n", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] %sNode integrity cannot be fully verified, tampering by service controller is possible %s\n", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
		has_flexver = 0;
		sleep(1);
	}
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ");
	fflush(stdout);

	// Retrieve remote FlexVer™ authentication results
	elapsed_time_milliseconds = 0;
	*verification_result = 0;
	if (!has_flexver) {
		flexver_verification_time_hint = 10;
	}
	while (1) {
		snprintf(post_request_data, 512, "{\"command\":\"REQUEST_OPERATION_RESULTS\",\"operation\":\"FLEXVER_AUTHENTICATION\",\"machineID\":\"%s\",\"nonce\":\"%s\"}", destination_machine_id, random_nonce_encoded);
		post_request_data[512] = 0;
		curl_easy_setopt(curl, CURLOPT_URL, control_server_api_url);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_request_data);
		mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending request '%s' to remote server...\n", post_request_data);
		mticlient_reset_curl_data_buffer();
		curl_ret = curl_easy_perform(curl);
		if (curl_ret != CURLE_OK) {
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to request FlexVer™ verification on remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
			return 1;
		}
		else {
			mticlient_process_received_data();
		}
		if (json_valid_api_response == 0) {
			// Break out of loop on hard error
			break;
		}
		if (json_valid_api_response == 1) {
			uint8_t operation_in_progress = 1;
			json_value = find_key_value(json_parser_start_token, "requestStatus");
			if (json_value && (strcmp(json_value, "VERIFICATION_COMPLETE") == 0)) {
				operation_in_progress = 0;
				*verification_result = 1;
			}
			if (json_value && (strcmp(json_value, "VERIFICATION_FAILED") == 0)) {
				operation_in_progress = 0;
				*verification_result = 0;
			}
			free(json_value);
			if (!operation_in_progress) {
				break;
			}
		}
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, ".");
		fflush(stdout);
		usleep(500000);
		elapsed_time_milliseconds += 500;
		if (elapsed_time_milliseconds > (flexver_verification_time_hint * 1000 * 3)) {
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			if (has_flexver) {
				mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] FlexVer™ verification timed out!\n");
			}
			else {
				mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] Key exchange timed out!\n");
			}
			return 1;
		}
	}
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "\r");
	fflush(stdout);

	if (*verification_result == 1) {
		*flexver_public_key = find_key_value(json_parser_start_token, "flexverPublicKey");
		*flexver_signature = find_key_value(json_parser_start_token, "flexverSignature");
		*flexver_status = find_key_value(json_parser_start_token, "flexverStatus");
		*flexver_internal_boot_count = find_key_value(json_parser_start_token, "flexverInternalBootCount");
		*flexver_system_firmware_hash = find_key_value(json_parser_start_token, "flexverSystemFirmwareHash");
		*flexver_mti_firmware_hash = find_key_value(json_parser_start_token, "flexverMTIFirmwareHash");
		*flexver_rsa_public_key = find_key_value(json_parser_start_token, "hostRSAPublicKey");

		if (!(*flexver_public_key) || !(*flexver_signature) || !(*flexver_status) || !(*flexver_internal_boot_count) || !(*flexver_system_firmware_hash) || !(*flexver_mti_firmware_hash) || !(*flexver_rsa_public_key)) {
			free(*flexver_public_key);
			free(*flexver_signature);
			free(*flexver_status);
			free(*flexver_internal_boot_count);
			free(*flexver_system_firmware_hash);
			free(*flexver_mti_firmware_hash);
			free(*flexver_rsa_public_key);
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to retrieve FlexVer™ verification results from remote server!\n");
			return 1;
		}

		*flexver_public_key_decoded = b64_decode(*flexver_public_key, strlen(*flexver_public_key));
		*flexver_signature_decoded = b64_decode(*flexver_signature, strlen(*flexver_signature));
		*flexver_status_raw = strtol(*flexver_status, NULL, 16);
		*flexver_internal_boot_count_raw = strtol(*flexver_internal_boot_count, NULL, 16);
		*flexver_system_firmware_hash_decoded = b64_decode(*flexver_system_firmware_hash, strlen(*flexver_system_firmware_hash));
		*flexver_mti_firmware_hash_decoded = b64_decode(*flexver_mti_firmware_hash, strlen(*flexver_mti_firmware_hash));
		*flexver_rsa_public_key_decoded = b64_decode(*flexver_rsa_public_key, strlen(*flexver_rsa_public_key));
		*flexver_public_key_string_1 = raw_data_to_hex_string(*flexver_public_key_decoded, 32);
		*flexver_public_key_string_2 = raw_data_to_hex_string((*flexver_public_key_decoded) + 32, FLEXVER_PUBLIC_KEY_LENGTH - 32);
		*flexver_system_firmware_hash_string = raw_data_to_hex_string(*flexver_system_firmware_hash_decoded, SYSTEM_FIRMWARE_HASH_LENGTH);
		*flexver_mti_firmware_hash_string = raw_data_to_hex_string(*flexver_mti_firmware_hash_decoded, MTI_FIRMWARE_HASH_LENGTH);

		verify_flexver_core_integrity_measurements(*flexver_status_raw, *flexver_internal_boot_count_raw, *flexver_public_key_decoded, *flexver_signature_decoded, *flexver_system_firmware_hash_decoded, *flexver_mti_firmware_hash_decoded, random_nonce_raw, *flexver_rsa_public_key_decoded, flexver_signature_valid, NULL, NULL);
	}

	if (!has_flexver) {
		*verification_result = 2;
	}

	return 0;
}

int main(int argc, char* argv[]) {
	char* json_value = NULL;
	char post_request_data[512+1];
	uint8_t random_nonce_raw[64];
	char* random_nonce_encoded = NULL;
	uint8_t flexver_signature_valid = 0;
	uint8_t flexver_verification_time_hint = 0;
	uint8_t verification_result = 0;
	uint8_t login_ready = 1;
	uint8_t has_nonflag_argument = 0;
	uint8_t has_username_argument = 0;
	uint8_t has_vm_id_argument = 0;
	uint8_t override_dest_node = 0;
	ssize_t index = -1;
	size_t max_arg_count = 0;
	size_t nonflag_argument_idx = 0;
	size_t dest_node_idx = 0;
	size_t username_idx = 0;
	size_t vm_id_idx = 0;
	char* remote_node_hostname = NULL;
	char* remote_node_port = NULL;
	char* remote_node_username = NULL;
	char* remote_node_password = NULL;
	int remote_node_port_numeric = 0;
	char* remote_node_hostname_known_hosts_format = NULL;
	char* remote_node_ip_addresses = NULL;
	char* flexver_public_key = NULL;
	char* flexver_system_firmware_hash = NULL;
	char* flexver_mti_firmware_hash = NULL;
	char* flexver_signature = NULL;
	char* flexver_status = NULL;
	char* flexver_internal_boot_count = NULL;
	char* flexver_rsa_public_key = NULL;
	uint8_t* flexver_public_key_decoded = NULL;
	uint8_t* flexver_signature_decoded = NULL;
	uint32_t flexver_status_raw = 0;
	uint16_t flexver_internal_boot_count_raw = 0;
	uint8_t* flexver_system_firmware_hash_decoded = NULL;
	uint8_t* flexver_mti_firmware_hash_decoded = NULL;
	uint8_t* flexver_rsa_public_key_decoded = NULL;
	char* flexver_public_key_string_1 = NULL;
	char* flexver_public_key_string_2 = NULL;
	char* flexver_system_firmware_hash_string = NULL;
	char* flexver_mti_firmware_hash_string = NULL;
	char** validated_flexver_public_key_list = NULL;
	size_t validated_flexver_public_key_list_count = 0;
	uint8_t** validated_system_firmware_hash_list = NULL;
	size_t validated_system_firmware_hash_list_count = 0;
	uint8_t** validated_mti_firmware_hash_list = NULL;
	size_t validated_mti_firmware_hash_list_count = 0;
	char* validated_flexver_public_key_tx_list = NULL;
	char* validated_system_firmware_hash_tx_list = NULL;
	char* validated_mti_firmware_hash_tx_list = NULL;
	size_t list_index = 0;
	FILE* file_access_test_fp = NULL;
	uint8_t flexver_confirmed_production_hash = 0;
	char flexver_firmware_hash_comments[1024+1];
	char* destination_machine_id = NULL;

	// Set debug level to INFO
	set_mticlient_loglevel(MTICLIENT_LOGLEVEL_INFO);

#ifdef DEBUG_LIBFLEXVER
	set_flexver_loglevel(FLEXVER_LOGLEVEL_INFO);
#else
	set_flexver_loglevel(FLEXVER_LOGLEVEL_NONE);
#endif

	// Show license and version information
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] MTIClient v" VERSION " Copyright © 2017 - 2018 Raptor Engineering, LLC.  All rights reserved.\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] WARNING: MTIClient may be covered under one or more patent(s) or patent(s) pending.\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Any use of this file outside of the context of access to Raptor Data Solutions, LLC's IntegriCloud™ services is strictly prohibited.\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] A single exception to this is made for the sole purpose of third-party security auditing of FlexVer™ and associated technologies.\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Please see the LICENSE file for additional details on the security audit exception.\n");
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");

	if (argc < 2) {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
		return 0;
	}

	// Parse incoming command arguments
	max_arg_count = 1;
	for (index = 1; index < argc; index++) {
		if (argv[index][0] == '-') {
			continue;
		}
		if (argv[index-1][0] == '-') {
			continue;
		}
		has_nonflag_argument = 1;
		nonflag_argument_idx = index;
		max_arg_count++;
		break;
	}
	if (command_argument_index(argv, argc, "-t") >= 0) {
		remote_hint_trust_client_keys = 1;
		max_arg_count += 1;
	}
	if ((index = command_argument_index(argv, argc, "-n")) >= 0) {
		if ((argc < (index + 2))
			|| (argv[index+1][0] == '-')) {
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Invalid number of arguments given to -n flag\n");
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
			return 1;
		}
		else {
			override_dest_node = 1;
			dest_node_idx = index + 1;
			max_arg_count += 2;
		}
	}
	if ((index = command_argument_index(argv, argc, "-u")) >= 0) {
		if ((argc < (index + 2))
			|| (argv[index+1][0] == '-')) {
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Invalid number of arguments given to -u flag\n");
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
			return 1;
		}
		else {
			has_username_argument = 1;
			username_idx = index + 1;
			max_arg_count += 2;
		}
	}
	if ((index = command_argument_index(argv, argc, "-i")) >= 0) {
		if ((argc < (index + 2))
			|| (argv[index+1][0] == '-')) {
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Invalid number of arguments given to -i flag\n");
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
			return 1;
		}
		else {
			has_vm_id_argument = 1;
			vm_id_idx = index + 1;
			max_arg_count += 2;
		}
	}
	if (argc > max_arg_count) {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
		return 0;
	}

	// Assemble file paths
	const char* home_directory = getenv("HOME");
	if (!home_directory) {
		home_directory = strdup("");
	}
	size_t production_fv_pk_list_file_1_expanded_length = 0;
	production_fv_pk_list_file_1_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_FV_PK_LIST_FILE_1);
	production_fv_pk_list_file_1_expanded_length++;
	production_fv_pk_list_file_1_expanded = malloc(production_fv_pk_list_file_1_expanded_length);
	if (!production_fv_pk_list_file_1_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_fv_pk_list_file_1_expanded, production_fv_pk_list_file_1_expanded_length, "%s%s", home_directory, PRODUCTION_FV_PK_LIST_FILE_1);
	production_fv_pk_list_file_1_expanded[production_fv_pk_list_file_1_expanded_length-1] = 0;
	size_t production_fv_pk_list_file_2_expanded_length = 0;
	production_fv_pk_list_file_2_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_FV_PK_LIST_FILE_2);
	production_fv_pk_list_file_2_expanded_length++;
	production_fv_pk_list_file_2_expanded = malloc(production_fv_pk_list_file_2_expanded_length);
	if (!production_fv_pk_list_file_2_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_fv_pk_list_file_2_expanded, production_fv_pk_list_file_2_expanded_length, "%s%s", home_directory, PRODUCTION_FV_PK_LIST_FILE_2);
	production_fv_pk_list_file_2_expanded[production_fv_pk_list_file_2_expanded_length-1] = 0;
	size_t production_sys_fw_list_file_1_expanded_length = 0;
	production_sys_fw_list_file_1_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_SYS_FW_LIST_FILE_1);
	production_sys_fw_list_file_1_expanded_length++;
	production_sys_fw_list_file_1_expanded = malloc(production_sys_fw_list_file_1_expanded_length);
	if (!production_sys_fw_list_file_1_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_sys_fw_list_file_1_expanded, production_sys_fw_list_file_1_expanded_length, "%s%s", home_directory, PRODUCTION_SYS_FW_LIST_FILE_1);
	production_sys_fw_list_file_1_expanded[production_sys_fw_list_file_1_expanded_length-1] = 0;
	size_t production_sys_fw_list_file_2_expanded_length = 0;
	production_sys_fw_list_file_2_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_SYS_FW_LIST_FILE_2);
	production_sys_fw_list_file_2_expanded_length++;
	production_sys_fw_list_file_2_expanded = malloc(production_sys_fw_list_file_2_expanded_length);
	if (!production_sys_fw_list_file_2_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_sys_fw_list_file_2_expanded, production_sys_fw_list_file_2_expanded_length, "%s%s", home_directory, PRODUCTION_SYS_FW_LIST_FILE_2);
	production_sys_fw_list_file_2_expanded[production_sys_fw_list_file_2_expanded_length-1] = 0;
	size_t production_mti_fw_list_file_1_expanded_length = 0;
	production_mti_fw_list_file_1_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_MTI_FW_LIST_FILE_1);
	production_mti_fw_list_file_1_expanded_length++;
	production_mti_fw_list_file_1_expanded = malloc(production_mti_fw_list_file_1_expanded_length);
	if (!production_mti_fw_list_file_1_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_mti_fw_list_file_1_expanded, production_mti_fw_list_file_1_expanded_length, "%s%s", home_directory, PRODUCTION_MTI_FW_LIST_FILE_1);
	production_mti_fw_list_file_1_expanded[production_mti_fw_list_file_1_expanded_length-1] = 0;
	size_t production_mti_fw_list_file_2_expanded_length = 0;
	production_mti_fw_list_file_2_expanded_length = snprintf(NULL, 0, "%s%s", home_directory, PRODUCTION_MTI_FW_LIST_FILE_2);
	production_mti_fw_list_file_2_expanded_length++;
	production_mti_fw_list_file_2_expanded = malloc(production_mti_fw_list_file_2_expanded_length);
	if (!production_mti_fw_list_file_2_expanded) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
		return 1;
	}
	snprintf(production_mti_fw_list_file_2_expanded, production_mti_fw_list_file_2_expanded_length, "%s%s", home_directory, PRODUCTION_MTI_FW_LIST_FILE_2);
	production_mti_fw_list_file_2_expanded[production_mti_fw_list_file_2_expanded_length-1] = 0;

	// Configure exit handler
	if (atexit(mticlient_cleanup) != 0) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to configure exit handler!\n");
		return 1;
	}

	// Initialise cURL and get global handle
	if (initialize_curl()) {
		return 1;
	}

	if (override_dest_node) {
		destination_machine_id = strdup(argv[dest_node_idx]);
	}
	else if (has_vm_id_argument && has_username_argument && !has_nonflag_argument) {
		// If VM ID given, look up associated node ID via API
		if (get_node_id_from_vm_info(argv[vm_id_idx], argv[username_idx], &destination_machine_id, 1)) {
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[INFO] Unable to retrieve node ID for given VM ID and username\n");
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
			return 1;
		}
	}
	else if (has_nonflag_argument && has_username_argument && !has_vm_id_argument) {
		// If VM name given, look up associated node ID via API
		if (get_node_id_from_vm_info(argv[nonflag_argument_idx], argv[username_idx], &destination_machine_id, 0)) {
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[INFO] Unable to retrieve node ID for given VM name and username\n");
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
			return 1;
		}
	}
	else {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] " MTISHELL_USAGE_STRING "\n");
		return 0;
	}
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Connecting to node ID %s\n", destination_machine_id);

	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Generating cryptographic nonce for verification operations...");
	uECC_RNG_Function fill_buffer_secure_random = uECC_get_rng();
	if (!fill_buffer_secure_random) {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "\n");
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] RNG access function not defined for this platform!\n");
		return 1;
	}
	if (!fill_buffer_secure_random(random_nonce_raw, 64)) {
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "\n");
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] Failed to generate cryptographic nonce!\n");
		return 1;
	}
	random_nonce_encoded = b64_encode(random_nonce_raw, 64);
	mticlient_log(MTICLIENT_LOGLEVEL_INFO, "done!\n");

	// Get remote FlexVer™ authentication time hint
	snprintf(post_request_data, 512, "{\"command\":\"REQUEST_PARAMETER\",\"parameter\":\"FLEXVER_VERIFICATION_TIME_HINT\",\"machineID\":\"%s\"}", destination_machine_id);
	post_request_data[512] = 0;
	curl_easy_setopt(curl, CURLOPT_URL, control_server_api_url);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_request_data);
	mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending request '%s' to remote server...\n", post_request_data);
	mticlient_reset_curl_data_buffer();
	curl_ret = curl_easy_perform(curl);
	if (curl_ret != CURLE_OK) {
		curl_easy_cleanup(curl);
		curl_global_cleanup();
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to obtain verification time hint from remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
		return 1;
	}
	else {
		mticlient_process_received_data();
	}
	json_value = find_key_value(json_parser_start_token, "verificationTimeHint");
	if (!json_value) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] Unable to obtain verification time hint from remote server!\n");
		return 1;
	}
	flexver_verification_time_hint = atoi(json_value);
	free(json_value);
	json_value = NULL;

	// Get remote node access data
	snprintf(post_request_data, 512, "{\"command\":\"REQUEST_PARAMETER\",\"parameter\":\"EXTERNAL_NODE_ACCESS_DATA\",\"machineID\":\"%s\"}", destination_machine_id);
	post_request_data[512] = 0;
	curl_easy_setopt(curl, CURLOPT_URL, control_server_api_url);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, post_request_data);
	mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending request '%s' to remote server...\n", post_request_data);
	mticlient_reset_curl_data_buffer();
	curl_ret = curl_easy_perform(curl);
	if (curl_ret != CURLE_OK) {
		curl_easy_cleanup(curl);
		curl_global_cleanup();
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to obtain node access data from remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
		return 1;
	}
	else {
		mticlient_process_received_data();
	}
	remote_node_hostname = find_key_value(json_parser_start_token, "hostname");
	remote_node_port = find_key_value(json_parser_start_token, "port");
	remote_node_username = find_key_value(json_parser_start_token, "username");
	remote_node_password = find_key_value(json_parser_start_token, "password");
	remote_node_port_numeric = atoi(remote_node_port);
	if (!remote_node_hostname || !remote_node_port || !remote_node_username || (remote_node_port_numeric <= 0)) {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "\n[ERR ] Unable to obtain node access data from remote server!\n");
		return 1;
	}

	// Read production system and MTI hashes from disk
	file_access_test_fp = fopen(production_sys_fw_list_file_1_expanded, "r");
	if (file_access_test_fp) {
		fclose(file_access_test_fp);
		read_hex_datalist_from_file(production_sys_fw_list_file_1_expanded, &validated_system_firmware_hash_list, &validated_system_firmware_hash_list_count, SYSTEM_FIRMWARE_HASH_LENGTH);
	}
	else {
		read_hex_datalist_from_file(production_sys_fw_list_file_1_expanded, &validated_system_firmware_hash_list, &validated_system_firmware_hash_list_count, SYSTEM_FIRMWARE_HASH_LENGTH);
	}
	file_access_test_fp = fopen(production_mti_fw_list_file_1_expanded, "r");
	if (file_access_test_fp) {
		fclose(file_access_test_fp);
		read_hex_datalist_from_file(production_mti_fw_list_file_1_expanded, &validated_mti_firmware_hash_list, &validated_mti_firmware_hash_list_count, MTI_FIRMWARE_HASH_LENGTH);
	}
	else {
		read_hex_datalist_from_file(production_mti_fw_list_file_2_expanded, &validated_mti_firmware_hash_list, &validated_mti_firmware_hash_list_count, MTI_FIRMWARE_HASH_LENGTH);
	}
	validated_system_firmware_hash_tx_list = write_raw_datalist_to_b64_string(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count, SYSTEM_FIRMWARE_HASH_LENGTH);
	validated_mti_firmware_hash_tx_list = write_raw_datalist_to_b64_string(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count, MTI_FIRMWARE_HASH_LENGTH);

	if (verify_remote_flexver_system(destination_machine_id, random_nonce_encoded, flexver_verification_time_hint, &verification_result, &flexver_public_key, &flexver_system_firmware_hash, &flexver_mti_firmware_hash, &flexver_signature, &flexver_status, &flexver_internal_boot_count, &flexver_rsa_public_key, &flexver_public_key_decoded, &flexver_signature_decoded, &flexver_status_raw, &flexver_internal_boot_count_raw, &flexver_system_firmware_hash_decoded, &flexver_mti_firmware_hash_decoded, &flexver_rsa_public_key_decoded, &flexver_public_key_string_1, &flexver_public_key_string_2, &flexver_system_firmware_hash_string, &flexver_mti_firmware_hash_string, random_nonce_raw, &flexver_signature_valid, 1)) {
		free(validated_system_firmware_hash_tx_list);
		free(validated_mti_firmware_hash_tx_list);
		free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
		free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
		validated_system_firmware_hash_list = NULL;
		validated_system_firmware_hash_list_count = 0;
		validated_mti_firmware_hash_list = NULL;
		validated_mti_firmware_hash_list_count = 0;
		return 1;
	}
	if ((verification_result == 0) || ((verification_result != 2) && (flexver_signature_valid == 0))) {
		// Clean up invalid result data
		free(flexver_public_key);
		free(flexver_signature);
		free(flexver_status);
		free(flexver_internal_boot_count);
		free(flexver_system_firmware_hash);
		free(flexver_mti_firmware_hash);
		free(flexver_rsa_public_key);
		free(flexver_public_key_decoded);
		free(flexver_signature_decoded);
		free(flexver_system_firmware_hash_decoded);
		free(flexver_mti_firmware_hash_decoded);
		free(flexver_rsa_public_key_decoded);
		free(flexver_public_key_string_1);
		free(flexver_public_key_string_2);
		free(flexver_system_firmware_hash_string);
		free(flexver_mti_firmware_hash_string);

		// Retry before reporting failure to user in case a transient communication issue was encountered
		if (verify_remote_flexver_system(destination_machine_id, random_nonce_encoded, flexver_verification_time_hint, &verification_result, &flexver_public_key, &flexver_system_firmware_hash, &flexver_mti_firmware_hash, &flexver_signature, &flexver_status, &flexver_internal_boot_count, &flexver_rsa_public_key, &flexver_public_key_decoded, &flexver_signature_decoded, &flexver_status_raw, &flexver_internal_boot_count_raw, &flexver_system_firmware_hash_decoded, &flexver_mti_firmware_hash_decoded, &flexver_rsa_public_key_decoded, &flexver_public_key_string_1, &flexver_public_key_string_2, &flexver_system_firmware_hash_string, &flexver_mti_firmware_hash_string, random_nonce_raw, &flexver_signature_valid, 2)) {
			free(validated_system_firmware_hash_tx_list);
			free(validated_mti_firmware_hash_tx_list);
			free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
			free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
			validated_system_firmware_hash_list = NULL;
			validated_system_firmware_hash_list_count = 0;
			validated_mti_firmware_hash_list = NULL;
			validated_mti_firmware_hash_list_count = 0;
			return 1;
		}
	}

	// Verify returned values
	if ((verification_result == 1) || (verification_result == 2)) {
		if (verification_result == 1) {
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Integrity verification complete!\n");
		}
		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");

		uint8_t audit_failure = 0;

		// Retrieve audit log from remote server
		curl_easy_setopt(curl, CURLOPT_URL, audit_log_url);
		curl_easy_setopt(curl, CURLOPT_HTTPGET, 1L);
		mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Sending audit log request to remote server...\n");
		mticlient_reset_curl_data_buffer();
		curl_ret = curl_easy_perform(curl);
		if (curl_ret != CURLE_OK) {
			free(flexver_public_key_decoded);
			free(flexver_signature_decoded);
			free(flexver_system_firmware_hash_decoded);
			free(flexver_mti_firmware_hash_decoded);
			free(flexver_rsa_public_key_decoded);
			free(flexver_public_key_string_1);
			free(flexver_public_key_string_2);
			free(flexver_system_firmware_hash_string);
			free(flexver_mti_firmware_hash_string);
			free(validated_system_firmware_hash_tx_list);
			free(validated_mti_firmware_hash_tx_list);
			free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
			free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
			validated_system_firmware_hash_list = NULL;
			validated_system_firmware_hash_list_count = 0;
			validated_mti_firmware_hash_list = NULL;
			validated_mti_firmware_hash_list_count = 0;
			free(flexver_public_key);
			free(flexver_signature);
			free(flexver_status);
			free(flexver_internal_boot_count);
			free(flexver_system_firmware_hash);
			free(flexver_mti_firmware_hash);
			free(flexver_rsa_public_key);
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to request FlexVer™ verification on remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
			return 1;
		}
		else {
			audit_failure = mticlient_process_audit_log(flexver_public_key, &validated_flexver_public_key_list, &validated_flexver_public_key_list_count);
			validated_flexver_public_key_tx_list = write_datalist_to_string(validated_flexver_public_key_list, validated_flexver_public_key_list_count);
			free_datalist(validated_flexver_public_key_list, validated_flexver_public_key_list_count);
		}

		free(flexver_signature);
		free(flexver_status);
		free(flexver_internal_boot_count);
		free(flexver_rsa_public_key);

		if (!verify_flexver_core_integrity_measurements(flexver_status_raw, flexver_internal_boot_count_raw, flexver_public_key_decoded, flexver_signature_decoded, flexver_system_firmware_hash_decoded, flexver_mti_firmware_hash_decoded, random_nonce_raw, flexver_rsa_public_key_decoded, &flexver_signature_valid, NULL, NULL)) {
			// Show remote system firmware information
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]            FlexVer™ public key:      [%s\n", flexver_public_key_string_1);
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]                                       %s]\n", flexver_public_key_string_2);
			mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");
			if (verification_result == 1) {
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] [ %sATTEST%s ] Origin: %s\n", CONSOLE_FONT_GREEN, CONSOLE_NORMAL, "FlexVer™");

				// System firmware
				flexver_confirmed_production_hash = 0;
				for (list_index = 0; list_index < validated_system_firmware_hash_list_count; list_index++) {
					if (memcmp(validated_system_firmware_hash_list[list_index], flexver_system_firmware_hash_decoded, SYSTEM_FIRMWARE_HASH_LENGTH) == 0) {
						flexver_confirmed_production_hash = 1;
					}
				}
				memset(flexver_firmware_hash_comments, 0, sizeof(flexver_firmware_hash_comments));
				if (flexver_confirmed_production_hash) {
					if (flexver_signature_valid && (!audit_failure)) {
						snprintf(flexver_firmware_hash_comments, 1024, " %sVERIFIED%s", CONSOLE_FONT_GREEN, CONSOLE_NORMAL);
					}
					else {
						snprintf(flexver_firmware_hash_comments, 1024, " %sUNKNOWN%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
					}
				}
				else {
					snprintf(flexver_firmware_hash_comments, 1024, " %sUNVERIFIED%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
				}
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]            System firmware hash:    %s\n[INFO]            System firmware status: %s\n", flexver_system_firmware_hash_string, flexver_firmware_hash_comments);

				// MTI firmware
				flexver_confirmed_production_hash = 0;
				for (list_index = 0; list_index < validated_mti_firmware_hash_list_count; list_index++) {
					if (memcmp(validated_mti_firmware_hash_list[list_index], flexver_mti_firmware_hash_decoded, MTI_FIRMWARE_HASH_LENGTH) == 0) {
						flexver_confirmed_production_hash = 1;
					}
				}
				memset(flexver_firmware_hash_comments, 0, sizeof(flexver_firmware_hash_comments));
				if (flexver_confirmed_production_hash) {
					if (flexver_signature_valid && (!audit_failure)) {
						snprintf(flexver_firmware_hash_comments, 1024, " %sVERIFIED%s", CONSOLE_FONT_GREEN, CONSOLE_NORMAL);
					}
					else {
						snprintf(flexver_firmware_hash_comments, 1024, " %sUNKNOWN%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
					}
				}
				else {
					snprintf(flexver_firmware_hash_comments, 1024, " %sUNVERIFIED%s", CONSOLE_FONT_YELLOW, CONSOLE_NORMAL);
				}
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO]            MTI firmware hash:       %s\n[INFO]            MTI firmware status:    %s\n", flexver_mti_firmware_hash_string, flexver_firmware_hash_comments);

				// Show FlexVer™ status results
				char* flexver_decoded_status_ii = NULL;
				char* flexver_decoded_status = decode_flexver_status_field(flexver_status_raw, flexver_signature_valid, audit_failure, 1, &flexver_decoded_status_ii);
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] %s\n", flexver_decoded_status);
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] -----------------------------------------------------------------------------------------------------------------------------------\n");
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] %s\n", flexver_decoded_status_ii);
				mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] ===================================================================================================================================\n");
				free(flexver_decoded_status);
				free(flexver_decoded_status_ii);
			}
		}
		else {
			free(flexver_public_key_decoded);
			free(flexver_signature_decoded);
			free(flexver_system_firmware_hash_decoded);
			free(flexver_mti_firmware_hash_decoded);
			free(flexver_rsa_public_key_decoded);
			free(flexver_public_key_string_1);
			free(flexver_public_key_string_2);
			free(flexver_system_firmware_hash_string);
			free(flexver_mti_firmware_hash_string);
			free(flexver_public_key);
			free(flexver_system_firmware_hash);
			free(flexver_mti_firmware_hash);
			free(validated_flexver_public_key_tx_list);
			free(validated_system_firmware_hash_tx_list);
			free(validated_mti_firmware_hash_tx_list);
			free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
			free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
			validated_system_firmware_hash_list = NULL;
			validated_system_firmware_hash_list_count = 0;
			validated_mti_firmware_hash_list = NULL;
			validated_mti_firmware_hash_list_count = 0;
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to parse FlexVer™ verification results from remote server!\n");
			return 1;
		}

		// Look up IP address(es) of remote node
		remote_node_ip_addresses = get_host_ip_addresses(remote_node_hostname, remote_node_port_numeric);

		// Format remote node hostname for SSH known hosts file
		if (remote_node_port_numeric == 22) {
			remote_node_hostname_known_hosts_format = strdup(remote_node_hostname);
		}
		else {
			size_t remote_node_hostname_known_hosts_format_length = 0;
			remote_node_hostname_known_hosts_format_length = snprintf(NULL, 0, "[%s]:%d", remote_node_hostname, remote_node_port_numeric);
			remote_node_hostname_known_hosts_format_length++;
			remote_node_hostname_known_hosts_format = malloc(remote_node_hostname_known_hosts_format_length);
			if (remote_node_hostname_known_hosts_format) {
				memset(remote_node_hostname_known_hosts_format, 0, remote_node_hostname_known_hosts_format_length);
				snprintf(remote_node_hostname_known_hosts_format, remote_node_hostname_known_hosts_format_length, "[%s]:%d", remote_node_hostname, remote_node_port_numeric);
				remote_node_hostname_known_hosts_format[remote_node_hostname_known_hosts_format_length-1] = 0;
			}
		}
		if (!remote_node_hostname_known_hosts_format) {
			free(flexver_public_key_decoded);
			free(flexver_signature_decoded);
			free(flexver_system_firmware_hash_decoded);
			free(flexver_mti_firmware_hash_decoded);
			free(flexver_rsa_public_key_decoded);
			free(flexver_public_key_string_1);
			free(flexver_public_key_string_2);
			free(flexver_system_firmware_hash_string);
			free(flexver_mti_firmware_hash_string);
			free(flexver_public_key);
			free(flexver_system_firmware_hash);
			free(flexver_mti_firmware_hash);
			free(validated_flexver_public_key_tx_list);
			free(validated_system_firmware_hash_tx_list);
			free(validated_mti_firmware_hash_tx_list);
			free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
			free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
			validated_system_firmware_hash_list = NULL;
			validated_system_firmware_hash_list_count = 0;
			validated_mti_firmware_hash_list = NULL;
			validated_mti_firmware_hash_list_count = 0;
			curl_easy_cleanup(curl);
			curl_global_cleanup();
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory!\n");
			return 1;
		}

		// Extract SSH public key components from FlexVer™ results
		char ssh_pubkey_modulus[512];
		char ssh_pubkey_exponent[16];
		uint32_t ssh_pubkey_modulus_length = 0;
		uint32_t ssh_pubkey_exponent_length = 0;
		memset(ssh_pubkey_modulus, 0, 512);
		memset(ssh_pubkey_exponent, 0, 16);
		extract_flexver_ssh_public_key_components(flexver_rsa_public_key_decoded, ssh_pubkey_exponent, ssh_pubkey_modulus, &ssh_pubkey_exponent_length, &ssh_pubkey_modulus_length);

		// Assemble encoded SSH public key for SSH known_hosts file
		char* ssh_pubkey_buffer = NULL;
		char* ssh_pubkey_buffer_encoded = 0;
		size_t ssh_rsa_pubkey_preamble_length = sizeof(ssh_rsa_pubkey_preamble) / sizeof(ssh_rsa_pubkey_preamble[0]);
		uint32_t ssh_pubkey_buffer_length = ssh_rsa_pubkey_preamble_length + ssh_pubkey_modulus_length + ssh_pubkey_exponent_length + 4 + 4;
		uint32_t ssh_pubkey_exponent_length_be = htonl(ssh_pubkey_exponent_length);
		uint32_t ssh_pubkey_modulus_length_be = htonl(ssh_pubkey_modulus_length);
		ssh_pubkey_buffer = malloc(ssh_pubkey_buffer_length);
		if (ssh_pubkey_buffer) {
			memset(ssh_pubkey_buffer, 0, ssh_pubkey_buffer_length);
			memcpy(ssh_pubkey_buffer, ssh_rsa_pubkey_preamble, ssh_rsa_pubkey_preamble_length);
			memcpy(ssh_pubkey_buffer + ssh_rsa_pubkey_preamble_length, &ssh_pubkey_exponent_length_be, 4);
			memcpy(ssh_pubkey_buffer + ssh_rsa_pubkey_preamble_length + 4, ssh_pubkey_exponent, ssh_pubkey_exponent_length);
			memcpy(ssh_pubkey_buffer + ssh_rsa_pubkey_preamble_length + 4 + ssh_pubkey_exponent_length, &ssh_pubkey_modulus_length_be, 4);
			memcpy(ssh_pubkey_buffer + ssh_rsa_pubkey_preamble_length + 4 + ssh_pubkey_exponent_length + 4, flexver_rsa_public_key_decoded, ssh_pubkey_modulus_length);
			ssh_pubkey_buffer_encoded = b64_encode(ssh_pubkey_buffer, ssh_pubkey_buffer_length);
			free(ssh_pubkey_buffer);

			char* known_hosts_buffer = NULL;
			size_t known_hosts_buffer_length = 0;
			known_hosts_buffer_length = snprintf(NULL, 0, SSH_KNOWN_HOSTS_FORMAT, remote_node_hostname_known_hosts_format, (remote_node_ip_addresses)?remote_node_ip_addresses:"", ssh_pubkey_buffer_encoded);
			known_hosts_buffer_length++;
			known_hosts_buffer = malloc(known_hosts_buffer_length);
			if (known_hosts_buffer) {
				memset(known_hosts_buffer, 0, known_hosts_buffer_length);
				snprintf(known_hosts_buffer, known_hosts_buffer_length, SSH_KNOWN_HOSTS_FORMAT, remote_node_hostname_known_hosts_format, (remote_node_ip_addresses)?remote_node_ip_addresses:"", ssh_pubkey_buffer_encoded);
				known_hosts_buffer[known_hosts_buffer_length-1] = 0;
				free(ssh_pubkey_buffer_encoded);

				// Create SSH known_hosts file
				ssh_known_hosts_fd = mkstemp(ssh_known_hosts_filename);
				if (ssh_known_hosts_fd >= 0) {
					// Write plaintext to file
					write(ssh_known_hosts_fd, known_hosts_buffer, known_hosts_buffer_length-1);
					fsync(ssh_known_hosts_fd);
					free(known_hosts_buffer);
				}
				else {
					mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to write SSH public key to temporary file!\n");
					free(known_hosts_buffer);
					login_ready = 0;
				}

				// Create SSH password file
				ssh_password_fd = mkstemp(ssh_password_filename);
				if (ssh_password_fd >= 0) {
					// Write password to file
					write(ssh_password_fd, remote_node_password, strlen(remote_node_password));
					fsync(ssh_password_fd);
				}
				else {
					mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to write SSH password to temporary file!\n");
					login_ready = 0;
				}
			}
			else {
				mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory for SSH known hosts creation!\n");
				free(ssh_pubkey_buffer_encoded);
				login_ready = 0;
			}
		}
		else {
			mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to allocate memory for SSH public key conversion!\n");
			login_ready = 0;
		}

		free(flexver_public_key_decoded);
		free(flexver_signature_decoded);
		free(flexver_system_firmware_hash_decoded);
		free(flexver_mti_firmware_hash_decoded);
		free(flexver_rsa_public_key_decoded);
		free(flexver_public_key_string_1);
		free(flexver_public_key_string_2);
		free(flexver_system_firmware_hash_string);
		free(flexver_mti_firmware_hash_string);
	}
	else {
		mticlient_log(MTICLIENT_LOGLEVEL_ERR, "[ERR ] Unable to obtain FlexVer™ verification results from remote server!  Detail: %s\n", curl_easy_strerror(curl_ret));
		login_ready = 0;
	}
	free(random_nonce_encoded);

	if ((flexver_signature_valid || (verification_result == 2)) && login_ready) {
		sanitize_command_buffer(remote_node_username);
		sanitize_command_buffer(remote_node_hostname);
		sanitize_command_buffer(remote_node_port);

		mticlient_log(MTICLIENT_LOGLEVEL_INFO, "[INFO] Connecting to remote server...\n");

		// Generate SSH environment
		char* ssh_environment_buffer = NULL;
		size_t ssh_environment_buffer_length = 0;
		ssh_environment_buffer_length = snprintf(NULL, 0, SSH_ENVIRONMENT_FORMAT, flexver_public_key, validated_flexver_public_key_tx_list, flexver_system_firmware_hash, validated_system_firmware_hash_tx_list, flexver_mti_firmware_hash, validated_mti_firmware_hash_tx_list, (remote_hint_trust_client_keys)?"TRUE":"FALSE");
		ssh_environment_buffer_length++;
		ssh_environment_buffer = malloc(ssh_environment_buffer_length);
		if (ssh_environment_buffer) {
			memset(ssh_environment_buffer, 0, ssh_environment_buffer_length);
			snprintf(ssh_environment_buffer, ssh_environment_buffer_length, SSH_ENVIRONMENT_FORMAT, flexver_public_key, validated_flexver_public_key_tx_list, flexver_system_firmware_hash, validated_system_firmware_hash_tx_list, flexver_mti_firmware_hash, validated_mti_firmware_hash_tx_list, (remote_hint_trust_client_keys)?"TRUE":"FALSE");
			ssh_environment_buffer[ssh_environment_buffer_length-1] = 0;
		}

		// Generate SSH command
		char* ssh_command_buffer = NULL;
		size_t ssh_command_buffer_length = 0;
		ssh_command_buffer_length = snprintf(NULL, 0, SSH_COMMAND_FORMAT, (ssh_environment_buffer)?ssh_environment_buffer:"", ssh_password_filename, ssh_known_hosts_filename, remote_node_port, remote_node_username, remote_node_hostname);
		ssh_command_buffer_length++;
		ssh_command_buffer = malloc(ssh_command_buffer_length);
		if (ssh_command_buffer) {
			memset(ssh_command_buffer, 0, ssh_command_buffer_length);
			snprintf(ssh_command_buffer, ssh_command_buffer_length, SSH_COMMAND_FORMAT, (ssh_environment_buffer)?ssh_environment_buffer:"", ssh_password_filename, ssh_known_hosts_filename, remote_node_port, remote_node_username, remote_node_hostname);
			ssh_command_buffer[ssh_command_buffer_length-1] = 0;

			// Execute SSH command
			mticlient_log(MTICLIENT_LOGLEVEL_DEBUG, "[DEBG] Executing command '%s'\n", ssh_command_buffer);
			system(ssh_command_buffer);
			free(ssh_command_buffer);
		}

		unlink(ssh_password_filename);
		close(ssh_password_fd);
		ssh_password_fd = -1;

		unlink(ssh_known_hosts_filename);
		close(ssh_known_hosts_fd);
		ssh_known_hosts_fd = -1;

		free(ssh_environment_buffer);
	}

	// Clean up
	free(validated_flexver_public_key_tx_list);
	free(validated_system_firmware_hash_tx_list);
	free(validated_mti_firmware_hash_tx_list);
	free_hex_datalist(validated_system_firmware_hash_list, validated_system_firmware_hash_list_count);
	free_hex_datalist(validated_mti_firmware_hash_list, validated_mti_firmware_hash_list_count);
	validated_system_firmware_hash_list = NULL;
	validated_system_firmware_hash_list_count = 0;
	validated_mti_firmware_hash_list = NULL;
	validated_mti_firmware_hash_list_count = 0;

	free(remote_node_ip_addresses);
	free(remote_node_hostname);
	free(remote_node_hostname_known_hosts_format);
	free(remote_node_port);
	free(remote_node_username);
	free(remote_node_password);

	free(flexver_public_key);
	free(flexver_system_firmware_hash);
	free(flexver_mti_firmware_hash);
	free(destination_machine_id);

	curl_easy_cleanup(curl);
	curl_global_cleanup();

	if (flexver_signature_valid && login_ready) {
		return 0;
	}
	else {
		return 1;
	}
}
